import os

API_ENDPOINT = "/storage/v1"
UPLOAD_API_ENDPOINT = "/upload/storage/v1"
BATCH_API_ENDPOINT = "/batch/storage/v1"
DOWNLOAD_API_ENDPOINT = "/download/storage/v1"

# pyfilesystem assumes OS fs within CWD as base
STORAGE_BASE = os.path.abspath(os.environ.get("STORAGE_BASE", "./"))
STORAGE_DIR = os.environ.get("STORAGE_DIR", ".cloudstorage")



import datetime
import os
from io import BytesIO
from tempfile import NamedTemporaryFile
from unittest import TestCase as BaseTestCase

import fs
import requests
from google.api_core.exceptions import BadRequest, Conflict, NotFound
from google.auth.credentials import AnonymousCredentials, Signing

from gcp_storage_emulator.server import create_server
from gcp_storage_emulator.settings import STORAGE_BASE, STORAGE_DIR


TEST_TEXT = os.path.join(os.path.dirname(os.path.abspath(__file__)), "test_text.txt")


class FakeSigningCredentials(Signing, AnonymousCredentials):
    def sign_bytes(self, message):
        return b"foobar"

    @property
    def signer_email(self):
        return "foobar@example.tld"

    @property
    def signer(self):
        pass


def _get_storage_client(http):
    """Gets a python storage client"""
    os.environ["STORAGE_EMULATOR_HOST"] = "http://localhost:9023"

    # Cloud Storage uses environment variables to configure API endpoints for
    # file upload - which is read at module import time
    from google.cloud import storage

    if os.getenv("DEBUG"):
        from http import client as http_client

        http_client.HTTPConnection.debuglevel = 5
    return storage.Client(
        project="[PROJECT]",
        _http=http,
        client_options={"api_endpoint": "http://localhost:9023"},
    )


class ServerBaseCase(BaseTestCase):
    @classmethod
    def setUpClass(cls):
        cls._server = create_server("localhost", 9023, in_memory=False)
        cls._server.start()

    @classmethod
    def tearDownClass(cls):
        cls._server.wipe()
        cls._server.stop()

    def setUp(self):
        self._session = requests.Session()
        self._client = _get_storage_client(self._session)
        self._server.wipe()


class BucketsTests(BaseTestCase):
    @classmethod
    def setUpClass(cls):
        cls._server = create_server("localhost", 9023, in_memory=False)
        cls._server.start()

    @classmethod
    def tearDownClass(cls):
        cls._server.wipe()
        cls._server.stop()

    def setUp(self):
        self._server.wipe()
        self._session = requests.Session()
        self._client = _get_storage_client(self._session)

    def test_bucket_creation(self):
        bucket = self._client.create_bucket("bucket_name")
        self.assertEqual(bucket.project_number, 1234)

    def test_bucket_creation_no_override(self):
        self._client.create_bucket("bucket_name")
        with self.assertRaises(Conflict):
            self._client.create_bucket("bucket_name")

    def test_bucket_list(self):
        bucket = self._client.create_bucket("bucket_name")
        all_bucket_names = [b.name for b in self._client.list_buckets()]
        self.assertIn(bucket.name, all_bucket_names)

    def test_bucket_get_existing(self):
        bucket = self._client.create_bucket("bucket_name")
        fetched_bucket = self._client.get_bucket("bucket_name")
        self.assertEqual(fetched_bucket.name, bucket.name)

    def test_bucket_get_existing_with_dot(self):
        bucket = self._client.create_bucket("bucket.name")
        fetched_bucket = self._client.get_bucket("bucket.name")
        self.assertEqual(fetched_bucket.name, bucket.name)

    def test_bucket_get_non_existing(self):
        with self.assertRaises(NotFound):
            self._client.get_bucket("bucket_name")

    def test_bucket_delete(self):
        bucket = self._client.create_bucket("bucket_name")
        bucket.delete()

        with self.assertRaises(NotFound):
            self._client.get_bucket("bucket_name")

    def test_bucket_delete_removes_file(self):
        bucket = self._client.create_bucket("bucket_name")
        bucket.delete()

        with fs.open_fs(os.path.join(STORAGE_BASE, STORAGE_DIR)) as pwd:
            self.assertFalse(pwd.exists("bucket_name"))

    def test_bucket_delete_non_existing(self):
        # client.bucket doesn't create the actual bucket resource remotely
        bucket = self._client.bucket("bucket_name")
        with self.assertRaises(NotFound):
            bucket.delete()

    def test_bucket_delete_non_empty(self):
        bucket = self._client.create_bucket("bucket_name")
        blob = bucket.blob("canttouchme.txt")
        blob.upload_from_string("This should prevent deletion if not force")

        with self.assertRaises(Conflict):
            bucket.delete()

        blob = bucket.get_blob("canttouchme.txt")
        self.assertIsNotNone(blob)

    def test_bucket_force_delete(self):
        bucket = self._client.create_bucket("bucket_name")
        blob = bucket.blob("cantouchme.txt")
        blob.upload_from_string("This should prevent deletion if not force")

        bucket.delete(force=True)

        blob = bucket.get_blob("cantouchme.txt")
        self.assertIsNone(blob)

        with fs.open_fs(os.path.join(STORAGE_BASE, STORAGE_DIR)) as pwd:
            self.assertFalse(pwd.exists("bucket_name"))


class DefaultBucketTests(BaseTestCase):
    def tearDown(self):
        if self._server:
            self._server.wipe()
            self._server.stop()
        return super().tearDown()

    def test_bucket_created(self):
        self._server = create_server(
            "localhost", 9023, in_memory=True, default_bucket="example.appspot.com"
        )
        self._server.start()
        self._session = requests.Session()
        self._client = _get_storage_client(self._session)
        bucket = self._client.get_bucket("example.appspot.com")
        self.assertEqual(bucket.name, "example.appspot.com")
        self.assertEqual(bucket.storage_class, "STANDARD")


class ObjectsTests(ServerBaseCase):
    def test_upload_from_string(self):
        content = "this is the content of the file\n"
        bucket = self._client.create_bucket("testbucket")
        blob = bucket.blob("testblob-name.txt")
        blob.upload_from_string(content)

        with fs.open_fs(os.path.join(STORAGE_BASE, STORAGE_DIR)) as pwd:
            read_content = pwd.readtext("testbucket/testblob-name.txt")
            self.assertEqual(read_content, content)

    def test_upload_from_text_file(self):
        bucket = self._client.create_bucket("testbucket")
        blob = bucket.blob("test_text.txt")
        with open(TEST_TEXT, "rb") as file:
            blob.upload_from_file(file)

            with fs.open_fs(os.path.join(STORAGE_BASE, STORAGE_DIR)) as pwd:
                read_content = pwd.readtext("testbucket/test_text.txt")

        with open(TEST_TEXT, "rb") as file:
            expected_content = str(file.read(), encoding="utf-8")
            self.assertEqual(read_content, expected_content)

    def test_upload_from_bin_file(self):
        test_binary = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "test_binary.png"
        )
        bucket = self._client.create_bucket("testbucket")
        blob = bucket.blob("binary.png")
        with open(test_binary, "rb") as file:
            blob.upload_from_file(file)

        with fs.open_fs(os.path.join(STORAGE_BASE, STORAGE_DIR)) as pwd:
            read_content = pwd.readbytes("testbucket/binary.png")

        with open(test_binary, "rb") as file:
            expected_content = file.read()
            self.assertEqual(read_content, expected_content)

    def test_upload_from_bin_file_cr_lf(self):
        content = b"\r\rheeeeei\r\n"
        test_binary = BytesIO(content)
        bucket = self._client.create_bucket("testbucket")
        blob = bucket.blob("binary_cr.png")

        blob.upload_from_file(test_binary, size=len(content))

        with fs.open_fs(os.path.join(STORAGE_BASE, STORAGE_DIR)) as pwd:
            read_content = pwd.readbytes("testbucket/binary_cr.png")

        self.assertEqual(read_content, content)

    def test_upload_from_file_name(self):
        test_binary = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "test_binary.png"
        )
        file_name = "test_binary.png"

        bucket = self._client.create_bucket("testbucket")
        blob = bucket.blob(file_name)
        blob.upload_from_filename(test_binary)
        blob = bucket.get_blob(file_name)
        with NamedTemporaryFile() as temp_file:
            blob.download_to_filename(temp_file.name)
            with open(test_binary, "rb") as orig_file:
                self.assertEqual(temp_file.read(), orig_file.read())

    def test_upload_from_file(self):
        test_binary = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "test_binary.png"
        )
        file_name = "test_binary.png"

        bucket = self._client.create_bucket("testbucket")
        blob = bucket.blob(file_name)
        with open(test_binary, "rb") as filehandle:
            blob.upload_from_file(filehandle)
            self.assertTrue(blob.id.startswith("testbucket/test_binary.png/"))

    def test_get(self):
        file_name = "testblob-name.txt"
        content = "this is the content of the file\n"
        bucket = self._client.create_bucket("testbucket")
        blob = bucket.blob(file_name)
        blob.upload_from_string(content)

        blob = bucket.get_blob(file_name)
        self.assertEqual(blob.name, file_name)

    def test_get_unicode(self):
        file_name = "tmp.ąćęłńóśźż.马铃薯.zip"
        content = "this is the content of the file\n"
        bucket = self._client.create_bucket("testbucket")
        blob = bucket.blob(file_name)
        blob.upload_from_string(content)

        blob = bucket.get_blob(file_name)
        self.assertEqual(blob.name, file_name)

    def test_get_nonexistent(self):
        bucket = self._client.create_bucket("testbucket")
        res = bucket.get_blob("idonotexist")

        self.assertIsNone(res)

        blob = bucket.blob("iexist")
        blob.upload_from_string("some_fake_content")
        res = bucket.get_blob("idonotexist")

        self.assertIsNone(res)

    def test_download_nonexistent(self):
        bucket = self._client.create_bucket("testbucket")
        blob = bucket.blob("idonotexist")
        with self.assertRaises(NotFound):
            blob.download_as_bytes()

    def test_upload_to_nonexistent_bucket(self):
        bucket = self._client.bucket("non-existent-test-bucket")
        blob = bucket.blob("idonotexisteither")
        with self.assertRaises(NotFound):
            blob.upload_from_string("some_content")

    def test_download_as_bytes(self):
        content = "The quick brown fox jumps over the lazy dog\n"
        bucket = self._client.create_bucket("testbucket")

        blob = bucket.blob("iexist")
        blob.upload_from_string(content)

        blob = bucket.get_blob("iexist")
        fetched_content = blob.download_as_bytes()
        self.assertEqual(fetched_content, content.encode("utf-8"))

    def test_download_range_start(self):
        content = b"123456789"
        bucket = self._client.create_bucket("testbucket")

        blob = bucket.blob("iexist")
        blob.upload_from_string(content)

        blob = bucket.get_blob("iexist")
        fetched_content = blob.download_as_bytes(start=2)
        self.assertEqual(fetched_content, b"3456789")

    def test_download_range_end(self):
        content = b"123456789"
        bucket = self._client.create_bucket("testbucket")

        blob = bucket.blob("iexist")
        blob.upload_from_string(content)

        blob = bucket.get_blob("iexist")
        fetched_content = blob.download_as_bytes(end=4)
        self.assertEqual(fetched_content, b"12345")

    def test_download_range_start_end(self):
        content = b"123456789"
        bucket = self._client.create_bucket("testbucket")

        blob = bucket.blob("iexist")
        blob.upload_from_string(content)

        blob = bucket.get_blob("iexist")
        fetched_content = blob.download_as_bytes(start=2, end=4)
        self.assertEqual(fetched_content, b"345")

    def test_set_content_encoding(self):
        content = "The quick brown fox jumps over the lazy dog\n"
        bucket = self._client.create_bucket("testbucket")

        blob = bucket.blob("testblob")
        blob.content_encoding = "gzip"
        blob.upload_from_string(content)
        blob.reload()
        self.assertEqual(blob.content_encoding, "gzip")

    def test_set_metadata(self):
        content = "The quick brown fox jumps over the lazy dog\n"
        bucket = self._client.create_bucket("testbucket")
        metadata = {"Color": "Pink"}

        blob = bucket.blob("testblob")
        blob.metadata = metadata
        blob.upload_from_string(content)
        blob.reload()
        self.assertEqual(blob.metadata, metadata)

    def test_set_custom_time(self):
        content = "The quick brown fox jumps over the lazy dog\n"
        bucket = self._client.create_bucket("testbucket")

        blob = bucket.blob("customtime")
        now = datetime.datetime.now(datetime.timezone.utc)
        blob.custom_time = now
        blob.upload_from_string(content)
        blob.reload()
        self.assertEqual(blob.custom_time, now)

    def test_patch_custom_time(self):
        content = "The quick brown fox jumps over the lazy dog\n"
        now = datetime.datetime.now(datetime.timezone.utc)
        bucket = self._client.create_bucket("testbucket")

        blob = bucket.blob("customtime")
        blob.upload_from_string(content)

        blob.reload()
        self.assertEqual(blob.custom_time, None)
        blob.custom_time = now
        blob.patch()
        blob.reload()
        self.assertEqual(blob.custom_time, now)

    def test_patch_custom_time_with_older_datetime(self):
        content = "The quick brown fox jumps over the lazy dog\n"
        newer = datetime.datetime.now(datetime.timezone.utc)
        older = datetime.datetime(2014, 11, 5, 20, 34, 37)
        bucket = self._client.create_bucket("testbucket")

        blob = bucket.blob("customtime")
        blob.upload_from_string(content)

        blob.reload()
        self.assertEqual(blob.custom_time, None)
        blob.custom_time = newer
        blob.patch()
        blob.reload()
        self.assertEqual(blob.custom_time, newer)
        blob.custom_time = older
        blob.patch()
        blob.reload()
        self.assertEqual(blob.custom_time, newer)

    def test_patch_content_encoding(self):
        content = "The quick brown fox jumps over the lazy dog\n"
        bucket = self._client.create_bucket("testbucket")

        blob = bucket.blob("testblob")
        blob.content_encoding = "gzip"
        blob.upload_from_string(content)
        blob.reload()
        metageneration = blob.metageneration
        self.assertEqual(blob.content_encoding, "gzip")
        blob.content_encoding = ""
        blob.patch()
        blob.reload()
        self.assertNotEqual(blob.metageneration, metageneration)
        self.assertEqual(blob.content_encoding, "")

    def test_valid_md5_hash(self):
        content = b"test"
        md5_hash = "CY9rzUYh03PK3k6DJie09g=="
        bucket = self._client.create_bucket("testbucket")
        blob = bucket.blob("hashtest")
        blob.md5_hash = md5_hash
        blob.upload_from_string(content)
        download_blob = bucket.get_blob("hashtest")
        self.assertEqual(download_blob.download_as_bytes(checksum="md5"), content)
        self.assertEqual(download_blob.md5_hash, md5_hash)

    def test_invalid_md5_hash(self):
        content = b"Hello World"
        bucket = self._client.create_bucket("testbucket")
        blob = bucket.blob("hashtest")
        blob.md5_hash = "deadbeef"
        with self.assertRaises(BadRequest):
            blob.upload_from_string(content)

    def test_valid_crc32c_hash(self):
        content = b"hello world"
        crc32c_hash = "yZRlqg=="
        bucket = self._client.create_bucket("testbucket")
        blob = bucket.blob("hashtest")
        blob.crc32c = crc32c_hash
        blob.upload_from_string(content)
        download_blob = bucket.get_blob("hashtest")
        self.assertEqual(download_blob.download_as_bytes(checksum="crc32c"), content)
        self.assertEqual(download_blob.crc32c, crc32c_hash)

    def test_invalid_crc32c_hash(self):
        content = b"Hello World"
        bucket = self._client.create_bucket("testbucket")
        blob = bucket.blob("hashtest")
        blob.crc32c = "deadbeef"
        with self.assertRaises(BadRequest):
            blob.upload_from_string(content)

    def test_download_binary_to_file(self):
        test_binary = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "test_binary.png"
        )
        bucket = self._client.create_bucket("testbucket")

        blob = bucket.blob("binary.png")
        with open(test_binary, "rb") as file:
            blob.upload_from_file(file, content_type="image/png")

        blob = bucket.get_blob("binary.png")
        fetched_file = BytesIO()
        blob.download_to_file(fetched_file)

        with open(test_binary, "rb") as file:
            self.assertEqual(fetched_file.getvalue(), file.read())

    def test_download_text_to_file(self):
        bucket = self._client.create_bucket("testbucket")

        blob = bucket.blob("text.txt")
        with open(TEST_TEXT, "rb") as file:
            blob.upload_from_file(file, content_type="text/plain; charset=utf-8")

        blob = bucket.get_blob("text.txt")
        fetched_file = BytesIO()
        blob.download_to_file(fetched_file)

        with open(TEST_TEXT, "rb") as file:
            self.assertEqual(fetched_file.getvalue(), file.read())

    def test_delete_object(self):
        bucket = self._client.create_bucket("bucket_name")
        blob = bucket.blob("canttouchme.txt")
        blob.upload_from_string("File content")

        with fs.open_fs(os.path.join(STORAGE_BASE, STORAGE_DIR)) as pwd:
            self.assertTrue(pwd.exists("bucket_name/canttouchme.txt"))
            blob.delete()

            self.assertIsNone(bucket.get_blob("cantouchme.txt"))
            self.assertFalse(pwd.exists("bucket_name/canttouchme.txt"))

    def test_delete_nonexistent_object(self):
        bucket = self._client.create_bucket("bucket_name")
        blob = bucket.blob("this-should-not-exists.txt")

        with self.assertRaises(NotFound):
            blob.delete()

    def test_create_within_directory(self):
        bucket = self._client.create_bucket("bucket_name")
        blob = bucket.blob("this/is/a/nested/file.txt")
        blob.upload_from_string("Not even joking!")

        with fs.open_fs(os.path.join(STORAGE_BASE, STORAGE_DIR)) as pwd:
            read_content = pwd.readtext("bucket_name/this/is/a/nested/file.txt")
            self.assertEqual(read_content, "Not even joking!")

    def test_create_within_multiple_time_does_not_break(self):
        bucket = self._client.create_bucket("bucket_name")
        blob = bucket.blob("this/is/a/nested/file.txt")
        blob.upload_from_string("Not even joking!")

        bucket.blob("this/is/another/nested/file.txt")
        blob.upload_from_string("Yet another one")

        with fs.open_fs(os.path.join(STORAGE_BASE, STORAGE_DIR)) as pwd:
            self.assertTrue(pwd.exists("bucket_name/this/is/a/nested/file.txt"))

    def _assert_blob_list(self, expected, actual):
        self.assertEqual([b.name for b in expected], [b.name for b in actual])

    def test_list_blobs_on_nonexistent_bucket(self):
        blobs = self._client.list_blobs("bucket_name")
        with self.assertRaises(NotFound):
            list(blobs)

    def test_list_blobs_on_empty_bucket(self):
        bucket = self._client.create_bucket("bucket_name")
        blobs = self._client.list_blobs(bucket)
        self._assert_blob_list(blobs, [])

    def test_list_blobs_on_entire_bucket(self):

        bucket_1 = self._client.create_bucket("bucket_name_1")
        bucket_2 = self._client.create_bucket("bucket_name_2")

        blob_1 = bucket_1.blob("a/b.txt")
        blob_1.upload_from_string("text")

        blob_2 = bucket_1.blob("c/d.txt")
        blob_2.upload_from_string("text")

        blob_3 = bucket_2.blob("a/b.txt")
        blob_3.upload_from_string("text")

        blobs = self._client.list_blobs(bucket_1)
        self._assert_blob_list(blobs, [blob_1, blob_2])

    def test_list_blobs_with_prefix(self):
        bucket = self._client.create_bucket("bucket_name")

        blob_1 = bucket.blob("a/b.txt")
        blob_1.upload_from_string("text")

        blob_2 = bucket.blob("a/b/c.txt")
        blob_2.upload_from_string("text")

        blob_3 = bucket.blob("b/c.txt")
        blob_3.upload_from_string("text")

        blobs = self._client.list_blobs(bucket, prefix="a")

        self._assert_blob_list(blobs, [blob_1, blob_2])

    def test_list_blobs_with_prefix_and_delimiter(self):
        bucket = self._client.create_bucket("bucket_name")

        blob_1 = bucket.blob("a/b.txt")
        blob_1.upload_from_string("text")

        blob_2 = bucket.blob("a/c.txt")
        blob_2.upload_from_string("text")

        blob_3 = bucket.blob("a/b/c.txt")
        blob_3.upload_from_string("text")

        blob_4 = bucket.blob("b/c.txt")
        blob_4.upload_from_string("text")

        blobs = self._client.list_blobs(bucket, prefix="a/", delimiter="/")

        self._assert_blob_list(blobs, [blob_1, blob_2])
        self.assertEqual(blobs.prefixes, {"a/b/"})

    def test_bucket_copy_existing(self):
        bucket = self._client.create_bucket("bucket_name")

        blob_1 = bucket.blob("a/b.txt")
        blob_1.upload_from_string("text")

        blob_2 = bucket.rename_blob(blob_1, "c/d.txt")

        blobs = self._client.list_blobs(bucket)
        self._assert_blob_list(blobs, [blob_2])

    def test_bucket_copy_non_existing(self):
        bucket = self._client.create_bucket("bucket_name")

        blob_1 = bucket.blob("a/b.txt")

        with self.assertRaises(NotFound):
            bucket.rename_blob(blob_1, "c/d.txt")

    def test_compose_create_new_blob(self):
        bucket = self._client.create_bucket("compose_test")
        data_1 = b"AAA\n"
        source_1 = bucket.blob("source-1")
        source_1.upload_from_string(data_1, content_type="text/plain")

        data_2 = b"BBB\n"
        source_2 = bucket.blob("source-2")
        source_2.upload_from_string(data_2, content_type="text/plain")

        destination = bucket.blob("destination")
        destination.content_type = "text/somethingelse"
        destination.compose([source_1, source_2])

        composed = destination.download_as_bytes()
        self.assertEqual(composed, data_1 + data_2)
        self.assertEqual(destination.content_type, "text/somethingelse")

    def test_compose_wo_content_type_set(self):
        bucket = self._client.create_bucket("compose_test")
        data_1 = b"AAA\n"
        source_1 = bucket.blob("source-1")
        source_1.upload_from_string(data_1, content_type="text/plain")

        data_2 = b"BBB\n"
        source_2 = bucket.blob("source-2")
        source_2.upload_from_string(data_2, content_type="text/plain")

        destination = bucket.blob("destination")
        destination.compose([source_1, source_2])

        composed = destination.download_as_bytes()
        self.assertEqual(composed, data_1 + data_2)
        self.assertEqual(destination.content_type, "text/plain")

    def test_compose_nonexistent(self):
        bucket = self._client.create_bucket("compose_test")
        source_1 = bucket.blob("source-1")
        source_2 = bucket.blob("source-2")

        destination = bucket.blob("destination")
        with self.assertRaises(NotFound):
            destination.compose([source_1, source_2])

    def test_batch_delete_one(self):
        content = "this is the content of the file\n"
        bucket = self._client.create_bucket("batchbucket")
        blob = bucket.blob("testblob-name1.txt")
        blob.upload_from_string(content)
        with self._client.batch():
            bucket.delete_blob("testblob-name1.txt")
        self.assertIsNone(bucket.get_blob("testblob-name1.txt"))

    def test_batch_delete_nonexistent_blob(self):
        bucket = self._client.create_bucket("batchbucket")
        with self.assertRaises(NotFound):
            with self._client.batch():
                bucket.delete_blob("does-not-exist.txt")

    def test_batch_patch_one(self):
        now = datetime.datetime.now(datetime.timezone.utc)
        content = "this is the content of the file\n"
        bucket = self._client.create_bucket("batchbucket")
        blob = bucket.blob("testblob-name1.txt")
        blob.upload_from_string(content)
        blob.reload()
        self.assertEqual(blob.custom_time, None)
        blob.custom_time = now
        with self._client.batch():
            blob.patch()
        blob = bucket.get_blob("testblob-name1.txt")
        self.assertEqual(blob.custom_time, now)

    def test_batch_delete_two(self):
        content = "this is the content of the file\n"
        bucket = self._client.create_bucket("batchbucket")
        blob = bucket.blob("testblob-name1.txt")
        blob.upload_from_string(content)
        blob = bucket.blob("testblob-name2.txt")
        blob.upload_from_string(content)
        with self._client.batch():
            bucket.delete_blob("testblob-name1.txt")
            bucket.delete_blob("testblob-name2.txt")
        self.assertIsNone(bucket.get_blob("testblob-name1.txt"))
        self.assertIsNone(bucket.get_blob("testblob-name2.txt"))

    def test_batch_patch_two(self):
        now = datetime.datetime.now(datetime.timezone.utc)
        content = "this is the content of the file\n"
        bucket = self._client.create_bucket("batchbucket")
        blob1 = bucket.blob("testblob-name1.txt")
        blob1.upload_from_string(content)
        blob2 = bucket.blob("testblob-name2.txt")
        blob2.upload_from_string(content)
        blob1.reload()
        blob2.reload()
        self.assertEqual(blob1.custom_time, None)
        self.assertEqual(blob2.custom_time, None)
        blob1.custom_time = now
        blob2.custom_time = now
        with self._client.batch():
            blob1.patch()
            blob2.patch()
        blob1 = bucket.get_blob("testblob-name1.txt")
        blob2 = bucket.get_blob("testblob-name2.txt")
        self.assertEqual(blob1.custom_time, now)
        self.assertEqual(blob2.custom_time, now)

    def test_batch_delete_patch(self):
        now = datetime.datetime.now(datetime.timezone.utc)
        content = "this is the content of the file\n"
        bucket = self._client.create_bucket("batchbucket")
        blob = bucket.blob("testblob-name1.txt")
        blob.upload_from_string(content)
        blob = bucket.blob("testblob-name2.txt")
        blob.upload_from_string(content)
        blob = bucket.blob("testblob-name3.txt")
        blob.upload_from_string(content)
        self.assertEqual(blob.custom_time, None)
        blob.custom_time = now
        with self._client.batch():
            bucket.delete_blob("testblob-name1.txt")
            bucket.delete_blob("testblob-name2.txt")
            blob.patch()
        self.assertIsNone(bucket.get_blob("testblob-name1.txt"))
        self.assertIsNone(bucket.get_blob("testblob-name2.txt"))
        blob = bucket.get_blob("testblob-name3.txt")
        self.assertEqual(blob.custom_time, now)

    def test_batch_delete_buckets(self):
        bucket1 = self._client.create_bucket("batchbucket1")
        bucket2 = self._client.create_bucket("batchbucket2")
        with self.assertRaises(NotFound):
            with self._client.batch():
                bucket1.delete()
                bucket1.delete()
                bucket2.delete()
        with self.assertRaises(NotFound):
            self._client.get_bucket("batchbucket1")
        with self.assertRaises(NotFound):
            self._client.get_bucket("batchbucket2")

    def test_resumable_upload_small_chunk_size(self):
        content = b"a" * 10000000
        bucket = self._client.create_bucket("testbucket")

        blob = bucket.blob("resumable-test", chunk_size=256 * 1024)
        blob.upload_from_string(content)

        blob = bucket.get_blob("resumable-test")
        fetched_content = blob.download_as_bytes()
        self.assertEqual(len(fetched_content), len(content))
        self.assertEqual(fetched_content, content)

    def test_resumable_upload_large_file(self):
        content = b"abcde12345" * 20000000
        bucket = self._client.create_bucket("testbucket")

        blob = bucket.blob("resumable-test")
        blob.upload_from_string(content)

        blob = bucket.get_blob("resumable-test")
        fetched_content = blob.download_as_bytes()
        self.assertEqual(len(fetched_content), len(content))
        self.assertEqual(fetched_content, content)

    def test_empty_blob(self):
        bucket = self._client.create_bucket("testbucket")
        bucket.blob("empty_blob").open("w").close()

        blob = bucket.get_blob("empty_blob")
        fetched_content = blob.download_as_bytes()
        self.assertEqual(fetched_content, b"")

    def test_signed_url_download(self):
        content = b"The quick brown fox jumps over the lazy dog"
        bucket = self._client.create_bucket("testbucket")

        blob = bucket.blob("signed-download")
        blob.upload_from_string(content, content_type="text/mycustom")

        url = blob.generate_signed_url(
            api_access_endpoint="http://localhost:9023",
            credentials=FakeSigningCredentials(),
            version="v4",
            expiration=datetime.timedelta(minutes=15),
            method="GET",
        )

        response = requests.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, content)
        self.assertEqual(response.headers["content-type"], "text/mycustom")

    def test_signed_url_upload(self):
        bucket = self._client.create_bucket("testbucket")
        blob = bucket.blob("signed-upload")
        url = blob.generate_signed_url(
            api_access_endpoint="http://localhost:9023",
            credentials=FakeSigningCredentials(),
            version="v4",
            expiration=datetime.timedelta(minutes=15),
            method="PUT",
        )
        with open(TEST_TEXT, "rb") as file:
            headers = {"Content-type": "text/plain"}
            response = requests.put(url, data=file, headers=headers)
            self.assertEqual(response.status_code, 200)

            blob_content = blob.download_as_bytes()
            file.seek(0)
            self.assertEqual(blob_content, file.read())
            self.assertEqual(blob.content_type, "text/plain")

    def test_signed_url_upload_to_nonexistent_bucket(self):
        bucket = self._client.bucket("non-existent-test-bucket")
        blob = bucket.blob("idonotexisteither")
        url = blob.generate_signed_url(
            api_access_endpoint="http://localhost:9023",
            credentials=FakeSigningCredentials(),
            version="v4",
            expiration=datetime.timedelta(minutes=15),
            method="PUT",
        )
        with open(TEST_TEXT, "rb") as file:
            response = requests.put(url, data=file)
            self.assertEqual(response.status_code, 404)

    def test_initiate_resumable_upload_without_metadata(self):
        url = "http://127.0.0.1:9023/upload/storage/v1/b/test_bucket/o?"
        url += "uploadType=resumable&name=test_file"
        self._client.create_bucket("test_bucket")
        headers = {"Content-type": "application/json"}
        response = requests.post(url, headers=headers)
        self.assertEqual(response.status_code, 200)

    def test_media_upload_without_metadata(self):
        url = "http://127.0.0.1:9023/upload/storage/v1/b/test_bucket/o?"
        url += "uploadType=media&name=test_file&contentEncoding=text%2Fplain"
        bucket = self._client.create_bucket("test_bucket")
        with open(TEST_TEXT, "rb") as file:
            headers = {"Content-type": "text/html"}
            response = requests.post(url, data=file, headers=headers)
            self.assertEqual(response.status_code, 200)
            blob = bucket.blob("test_file")
            blob_content = blob.download_as_bytes()
            file.seek(0)
            self.assertEqual(blob_content, file.read())
            self.assertEqual(blob.content_type, "text/plain")

    def test_upload_from_file_content_type_json(self):
        file_name = "test.json"
        content = b'[{"a": 1}]'
        bucket = self._client.create_bucket("testbucket")
        blob = bucket.blob(file_name)

        with NamedTemporaryFile() as temp_file:
            temp_file.write(content)
            temp_file.flush()
            temp_file.seek(0)
            blob.upload_from_file(temp_file, content_type="application/json")

        blob = bucket.get_blob(file_name)
        self.assertEqual(blob.name, file_name)
        self.assertEqual(blob.download_as_bytes(), content)


class HttpEndpointsTest(ServerBaseCase):
    """Tests for the HTTP endpoints defined by server.HANDLERS."""

    def _url(self, path):
        return os.environ["STORAGE_EMULATOR_HOST"] + path

    def test_download_by_url(self):
        """Objects should be downloadable over HTTP from the emulator client."""
        content = "Here is some content"
        bucket = self._client.create_bucket("anotherbucket")
        blob = bucket.blob("something.txt")
        blob.upload_from_string(content)

        url = self._url("/anotherbucket/something.txt")
        response = requests.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, content.encode("utf-8"))

    def test_download_by_dl_api_url(self):
        """Objects should be downloadable over HTTP from the emulator client."""
        content = "Here is some content 123"
        bucket = self._client.create_bucket("bucket")
        blob = bucket.blob("something.txt")
        blob.upload_from_string(content)

        url = self._url("/download/storage/v1/b/bucket/o/something.txt")
        response = requests.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, content.encode("utf-8"))

    def test_download_by_api_media_url(self):
        """Objects should be downloadable over HTTP from the emulator client."""
        content = "Here is some content 456"
        bucket = self._client.create_bucket("bucket")
        blob = bucket.blob("something.txt")
        blob.upload_from_string(content)

        url = self._url("/storage/v1/b/bucket/o/something.txt")
        response = requests.get(url, params={"alt": "media"})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, content.encode("utf-8"))

    def test_download_file_within_folder(self):
        """Cloud Storage allows folders within buckets, so the download URL should allow for this."""
        content = "Here is some content"
        bucket = self._client.create_bucket("yetanotherbucket")
        blob = bucket.blob("folder/contain~ing/something~v-1.0.α.txt")
        blob.upload_from_string(content)

        url = self._url("/yetanotherbucket/folder/contain~ing/something~v-1.0.α.txt")
        response = requests.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, content.encode("utf-8"))

    def test_wipe(self):
        """Objects should wipe the data"""
        storage_path = os.path.join(STORAGE_BASE, STORAGE_DIR)
        content = "Here is some content"
        bucket = self._client.create_bucket("anotherbucket1")
        blob = bucket.blob("something.txt")
        blob.upload_from_string(content)

        url = self._url("/wipe")
        response = requests.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertTrue(len(os.listdir(storage_path)) == 0)

    def test_wipe_keep_buckets(self):
        """Objects should wipe the data but keep the root buckets"""
        blob_path = "something.txt"
        bucket_name = "anewone"
        content = "Here is some content"
        bucket = self._client.create_bucket(bucket_name)
        blob = bucket.blob(blob_path)
        blob.upload_from_string(content)

        url = self._url("/wipe?keep-buckets=true")
        response = requests.get(url)
        self.assertEqual(response.status_code, 200)

        fetched_bucket = self._client.get_bucket(bucket_name)
        self.assertEqual(fetched_bucket.name, bucket.name)
        with self.assertRaises(NotFound):
            fetched_bucket.blob(blob_path).download_as_text()



import logging
from datetime import datetime, timezone
from http import HTTPStatus

from gcp_storage_emulator import settings
from gcp_storage_emulator.exceptions import Conflict, NotFound

logger = logging.getLogger("api.bucket")

CONFLICT = {
    "error": {
        "errors": [
            {
                "domain": "global",
                "reason": "conflict",
                "message": "You already own this bucket. Please select another name.",
            }
        ],
        "code": 409,
        "message": "You already own this bucket. Please select another name.",
    }
}

BAD_REQUEST = {
    "error": {
        "errors": [
            {"domain": "global", "reason": "invalid", "message": "Empty bucket name"}
        ],
        "code": 400,
        "message": "Empty bucket name",
    }
}


def _make_bucket_resource(bucket_name):
    now = datetime.now(timezone.utc).strftime("%Y-%m-%dT%H:%M:%S.%fZ")
    return {
        "kind": "storage#bucket",
        "id": bucket_name,
        "selfLink": "{}/b/{}".format(settings.API_ENDPOINT, bucket_name),
        "projectNumber": "1234",
        "name": bucket_name,
        "timeCreated": now,
        "updated": now,
        "metageneration": "1",
        "iamConfiguration": {
            "bucketPolicyOnly": {"enabled": False},
            "uniformBucketLevelAccess": {"enabled": False},
        },
        "location": "US",
        "locationType": "multi-region",
        "storageClass": "STANDARD",
        "etag": "CAE=",
    }


def get(request, response, storage, *args, **kwargs):
    name = request.params.get("bucket_name")
    if name and storage.buckets.get(name):
        response.json(storage.buckets.get(name))
    else:
        response.status = HTTPStatus.NOT_FOUND


def ls(request, response, storage, *args, **kwargs):
    logger.info("[BUCKETS] List received")
    response.json(
        {
            "kind": "storage#buckets",
            "items": list(storage.buckets.values()),
        }
    )


def create_bucket(name, storage):
    if storage.get_bucket(name):
        return False
    else:
        bucket = _make_bucket_resource(name)
        storage.create_bucket(name, bucket)
        return bucket


def insert(request, response, storage, *args, **kwargs):
    name = request.data.get("name")
    if name:
        logger.debug(
            "[BUCKETS] Received request to create bucket with name {}".format(name)
        )
        bucket = create_bucket(name, storage)
        if not bucket:
            response.status = HTTPStatus.CONFLICT
            response.json(CONFLICT)
        else:
            bucket = _make_bucket_resource(name)
            storage.create_bucket(name, bucket)
            response.json(bucket)
    else:
        response.status = HTTPStatus.BAD_REQUEST
        response.json(BAD_REQUEST)


def delete(request, response, storage, *args, **kwargs):
    name = request.params.get("bucket_name")
    if not name:
        response.status = HTTPStatus.BAD_REQUEST
        return response.json(BAD_REQUEST)

    try:
        storage.delete_bucket(name)
    except NotFound:
        response.status = HTTPStatus.NOT_FOUND
    except Conflict:
        response.status = HTTPStatus.CONFLICT



import hashlib
import json
import logging
import math
import re
import secrets
import string
import time
import urllib.parse
from base64 import b64encode
from copy import deepcopy
from datetime import datetime, timezone
from enum import IntEnum
from http import HTTPStatus

import google_crc32c

from gcp_storage_emulator.exceptions import Conflict, NotFound

logger = logging.getLogger("api.object")

_WRITABLE_FIELDS = (
    "cacheControl",
    "contentDisposition",
    "contentEncoding",
    "contentLanguage",
    "contentType",
    "crc32c",
    "customTime",
    "md5Hash",
    "metadata",
    "storageClass",
)

_HASH_HEADER = "X-Goog-Hash"

BAD_REQUEST = {
    "error": {
        "errors": [{"domain": "global", "reason": "invalid", "message": None}],
        "code": 400,
        "message": None,
    }
}

NOT_FOUND = {
    "error": {
        "errors": [{"domain": "global", "reason": "notFound", "message": None}],
        "code": 404,
        "message": None,
    }
}


MD5_CHECKSUM_ERROR = 'Provided MD5 hash "{}" doesn\'t match calculated MD5 hash "{}".'
CRC32C_CHECKSUM_ERROR = 'Provided CRC32C "{}" doesn\'t match calculated CRC32C "{}".'


class GoogleHTTPStatus(IntEnum):
    def __new__(cls, value, phrase, description=""):
        obj = int.__new__(cls, value)
        obj._value_ = value

        obj.phrase = phrase
        obj.description = description
        return obj

    RESUME_INCOMPLETE = 308, "Resume Incomplete"


def _handle_conflict(response, err):
    msg = str(err)
    response.status = HTTPStatus.BAD_REQUEST
    resp = deepcopy(BAD_REQUEST)
    resp["error"]["message"] = msg
    resp["error"]["errors"][0]["message"] = msg
    response.json(resp)


def _crc32c(content):
    if isinstance(content, str):
        content = content.encode()
    val = google_crc32c.Checksum(content)
    return b64encode(val.digest()).decode("ascii")


def _md5(content):
    if isinstance(content, str):
        content = content.encode()
    return b64encode(hashlib.md5(content).digest()).decode("ascii")


def _checksums(content, file_obj):
    crc32c_hash = _crc32c(content)
    obj_crc32c = file_obj.get("crc32c")
    md5_hash = _md5(content)
    obj_md5 = file_obj.get("md5Hash")
    if not obj_crc32c:
        file_obj["crc32c"] = crc32c_hash
    else:
        if obj_crc32c != crc32c_hash:
            raise Conflict(CRC32C_CHECKSUM_ERROR.format(obj_crc32c, crc32c_hash))
    if not obj_md5:
        file_obj["md5Hash"] = md5_hash
    else:
        if obj_md5 != md5_hash:
            raise Conflict(MD5_CHECKSUM_ERROR.format(obj_md5, md5_hash))
    if not file_obj.get("etag"):
        file_obj["etag"] = md5_hash
    return file_obj


def _patch_object(obj, metadata):
    if metadata:
        obj["metageneration"] = str(int(obj["metageneration"]) + 1)
        for key in _WRITABLE_FIELDS:
            val = metadata.get(key)
            if val is not None:
                if key == "customTime" and obj.get(key) and obj.get(key) > val:
                    continue
                obj[key] = val
    return obj


def _make_object_resource(
    base_url, bucket_name, object_name, content_type, content_length, metadata=None
):
    time_id = math.floor(time.time())
    now = datetime.now(timezone.utc).strftime("%Y-%m-%dT%H:%M:%S.%fZ")

    obj = {
        "kind": "storage#object",
        "id": "{}/{}/{}".format(bucket_name, object_name, time_id),
        "selfLink": "/storage/v1/b/{}/o/{}".format(bucket_name, object_name),
        "name": object_name,
        "bucket": bucket_name,
        "generation": str(time_id),
        "metageneration": "1",
        "contentType": content_type,
        "timeCreated": now,
        "updated": now,
        "storageClass": "STANDARD",
        "timeStorageClassUpdated": now,
        "size": content_length,
        "md5Hash": None,
        "mediaLink": "{}/download/storage/v1/b/{}/o/{}?generation={}&alt=media".format(
            base_url,
            bucket_name,
            object_name,
            time_id,
        ),
        "crc32c": None,
        "etag": None,
    }
    obj = _patch_object(obj, metadata)
    return obj


def _content_type_from_request(request, default=None):
    if "contentEncoding" in request.query:
        return request.query["contentEncoding"][0]
    return default


def _media_upload(request, response, storage):
    object_id = request.query["name"][0]
    content_type = _content_type_from_request(
        request, request.get_header("content-type")
    )
    obj = _make_object_resource(
        request.base_url,
        request.params["bucket_name"],
        object_id,
        content_type,
        str(len(request.data)),
    )
    obj = _checksums(request.data, obj)
    storage.create_file(
        request.params["bucket_name"],
        object_id,
        request.data,
        obj,
    )

    response.json(obj)


def _multipart_upload(request, response, storage):
    object_id = request.data["meta"].get("name")
    # Overrides the object metadata's name value, if any.
    if "name" in request.query:
        object_id = request.query["name"][0]
    content_type = _content_type_from_request(request, request.data["content-type"])
    obj = _make_object_resource(
        request.base_url,
        request.params["bucket_name"],
        object_id,
        content_type,
        str(len(request.data["content"])),
        request.data["meta"],
    )
    obj = _checksums(request.data["content"], obj)
    storage.create_file(
        request.params["bucket_name"],
        object_id,
        request.data["content"],
        obj,
    )

    response.json(obj)


def _create_resumable_upload(request, response, storage):
    # Workaround for libraries using POST method when they should be using PUT.
    if "upload_id" in request.query:
        return upload_partial(request, response, storage)
    if request.data:
        object_id = request.data.get("name")
    # Overrides the object metadata's name value, if any.
    if "name" in request.query:
        object_id = request.query["name"][0]
    content_type = _content_type_from_request(
        request, request.get_header("x-upload-content-type", "application/octet-stream")
    )
    content_length = request.get_header("x-upload-content-length", None)
    obj = _make_object_resource(
        request.base_url,
        request.params["bucket_name"],
        object_id,
        content_type,
        content_length,
    )
    id = storage.create_resumable_upload(
        request.params["bucket_name"],
        object_id,
        obj,
    )
    encoded_id = urllib.parse.urlencode(
        {
            "upload_id": id,
        }
    )
    response["Location"] = request.full_url + "&{}".format(encoded_id)


def _delete(storage, bucket_name, object_id):
    try:
        storage.delete_file(bucket_name, object_id)
        return True
    except NotFound:
        return False


def _patch(storage, bucket_name, object_id, metadata):
    try:
        obj = storage.get_file_obj(bucket_name, object_id)
        obj = _patch_object(obj, metadata)
        storage.patch_object(bucket_name, object_id, obj)
        return obj
    except NotFound:
        logger.error(
            "Could not patch {}/{}: with {}".format(bucket_name, object_id, metadata)
        )
        return None


def xml_upload(request, response, storage, *args, **kwargs):
    content_type = request.get_header("Content-Type", "application/octet-stream")
    obj = _make_object_resource(
        request.base_url,
        request.params["bucket_name"],
        request.params["object_id"],
        content_type,
        str(len(request.data)),
    )
    try:
        obj = _checksums(request.data, obj)
        storage.create_file(
            request.params["bucket_name"],
            request.params["object_id"],
            request.data,
            obj,
        )

    except NotFound:
        response.status = HTTPStatus.NOT_FOUND


def insert(request, response, storage, *args, **kwargs):
    uploadType = request.query.get("uploadType")

    if not uploadType or len(uploadType) == 0:
        response.status = HTTPStatus.BAD_REQUEST
        return

    uploadType = uploadType[0]

    try:
        if uploadType == "media":
            return _media_upload(request, response, storage)

        if uploadType == "resumable":
            return _create_resumable_upload(request, response, storage)

        if uploadType == "multipart":
            return _multipart_upload(request, response, storage)
    except NotFound:
        response.status = HTTPStatus.NOT_FOUND
    except Conflict as err:
        _handle_conflict(response, err)


def upload_partial(request, response, storage, *args, **kwargs):
    """https://cloud.google.com/storage/docs/performing-resumable-uploads"""
    upload_id = request.query.get("upload_id")[0]
    regex = r"^\s*bytes (?P<start>[0-9]+)-(?P<end>[0-9]+)/(?P<total_size>[0-9]+)$"
    pattern = re.compile(regex)
    content_range = request.get_header("Content-Range", "")
    match = pattern.fullmatch(content_range)
    try:
        obj = storage.get_resumable_file_obj(upload_id)
        if match:
            m_dict = match.groupdict()
            total_size = int(m_dict["total_size"])
            data = storage.add_to_resumable_upload(upload_id, request.data, total_size)
            if data is None:
                response.status = GoogleHTTPStatus.RESUME_INCOMPLETE
                response["Range"] = "bytes=0-{}".format(m_dict["end"])
                return
        else:
            data = request.data or b""

        obj = _checksums(data, obj)
        obj["size"] = str(len(data))
        storage.create_file(obj["bucket"], obj["name"], data, obj, upload_id)
        response.json(obj)
    except NotFound:
        response.status = HTTPStatus.NOT_FOUND
    except Conflict as err:
        _handle_conflict(response, err)


def get(request, response, storage, *args, **kwargs):
    if request.query.get("alt") and request.query.get("alt")[0] == "media":
        return download(request, response, storage)
    try:
        obj = storage.get_file_obj(
            request.params["bucket_name"], request.params["object_id"]
        )
        response.json(obj)
    except NotFound:
        response.status = HTTPStatus.NOT_FOUND


def ls(request, response, storage, *args, **kwargs):
    bucket_name = request.params["bucket_name"]
    prefix = request.query.get("prefix")[0] if request.query.get("prefix") else None
    delimiter = (
        request.query.get("delimiter")[0] if request.query.get("delimiter") else None
    )
    try:
        files, prefixes = storage.get_file_list(bucket_name, prefix, delimiter)
    except NotFound:
        response.status = HTTPStatus.NOT_FOUND
    else:
        response.json({"kind": "storage#object", "prefixes": prefixes, "items": files})


def copy(request, response, storage, *args, **kwargs):
    try:
        obj = storage.get_file_obj(
            request.params["bucket_name"], request.params["object_id"]
        )
    except NotFound:
        response.status = HTTPStatus.NOT_FOUND
        return

    dest_obj = _make_object_resource(
        request.base_url,
        request.params["dest_bucket_name"],
        request.params["dest_object_id"],
        obj["contentType"],
        obj["size"],
        obj,
    )

    file = storage.get_file(request.params["bucket_name"], request.params["object_id"])
    try:
        dest_obj = _checksums(file, dest_obj)
        storage.create_file(
            request.params["dest_bucket_name"],
            request.params["dest_object_id"],
            file,
            dest_obj,
        )
        response.json(dest_obj)
    except NotFound:
        response.status = HTTPStatus.NOT_FOUND
    except Conflict as err:
        _handle_conflict(response, err)


def compose(request, response, storage, *args, **kwargs):
    content_type = None
    dest_file = b""
    try:
        dest_properties = request.data["destination"]
        for src_obj in request.data["sourceObjects"]:
            if content_type is None:
                temp = storage.get_file_obj(
                    request.params["bucket_name"], src_obj["name"]
                )
                content_type = temp["contentType"]
            dest_file += storage.get_file(
                request.params["bucket_name"], src_obj["name"]
            )

    except NotFound:
        response.status = HTTPStatus.NOT_FOUND
        return

    dest_obj = _make_object_resource(
        request.base_url,
        request.params["bucket_name"],
        request.params["object_id"],
        content_type,
        len(dest_file),
        dest_properties,
    )

    try:
        dest_obj = _checksums(dest_file, dest_obj)
        storage.create_file(
            request.params["bucket_name"],
            request.params["object_id"],
            dest_file,
            dest_obj,
        )
        response.json(dest_obj)
    except NotFound:
        response.status = HTTPStatus.NOT_FOUND
    except Conflict as err:
        _handle_conflict(response, err)


def download(request, response, storage, *args, **kwargs):
    try:
        file = storage.get_file(
            request.params["bucket_name"], request.params["object_id"]
        )
        obj = storage.get_file_obj(
            request.params["bucket_name"], request.params["object_id"]
        )
        range = request.get_header("range", None)
        if range:
            regex = r"^\s*bytes=(?P<start>[0-9]+)-(?P<end>[0-9]*)$"
            pattern = re.compile(regex)
            match = pattern.fullmatch(range)
            if match:
                end = orig_len = len(file)
                m_dict = match.groupdict()
                start = int(m_dict["start"])
                if m_dict["end"]:
                    end = min(orig_len, int(m_dict["end"]) + 1)
                file = file[start:end]
                end -= 1
                response["Content-Range"] = "bytes {}-{}/{}".format(
                    start, end, orig_len
                )
                response.status = HTTPStatus.PARTIAL_CONTENT
        else:
            hash_header = "crc32c={},md5={}".format(obj["crc32c"], obj["md5Hash"])
            response[_HASH_HEADER] = hash_header

        response.write_file(file, content_type=obj.get("contentType"))
    except NotFound:
        response.status = HTTPStatus.NOT_FOUND


def delete(request, response, storage, *args, **kwargs):
    if not _delete(storage, request.params["bucket_name"], request.params["object_id"]):
        response.status = HTTPStatus.NOT_FOUND


def patch(request, response, storage, *args, **kwargs):
    obj = _patch(
        storage,
        request.params["bucket_name"],
        request.params["object_id"],
        request.data,
    )
    if obj:
        response.json(obj)
    else:
        response.status = HTTPStatus.NOT_FOUND


def batch(request, response, storage, *args, **kwargs):
    boundary = "batch_" + "".join(
        secrets.choice(string.ascii_uppercase + string.ascii_lowercase + string.digits)
        for _ in range(32)
    )
    response["Content-Type"] = "multipart/mixed; boundary={}".format(boundary)
    for item in request.data:
        resp_data = None
        response.write("--{}\r\nContent-Type: application/http\r\n".format(boundary))
        method = item.get("method")
        bucket_name = item.get("bucket_name")
        object_id = item.get("object_id")
        meta = item.get("meta")
        if method == "PATCH":
            resp_data = _patch(storage, bucket_name, object_id, meta)
            if resp_data:
                response.write("HTTP/1.1 200 OK\r\n")
                response.write("Content-Type: application/json; charset=UTF-8\r\n")
                response.write(json.dumps(resp_data))
                response.write("\r\n\r\n")
        if method == "DELETE":
            if object_id:
                resp_data = _delete(storage, bucket_name, object_id)
            else:
                try:
                    storage.delete_bucket(bucket_name)
                    resp_data = True
                except (Conflict, NotFound):
                    pass
            if resp_data:
                response.write("HTTP/1.1 204 No Content\r\n")
                response.write("Content-Type: application/json; charset=UTF-8\r\n")
        if not resp_data:
            msg = "No such object: {}/{}".format(bucket_name, object_id)
            resp_data = deepcopy(NOT_FOUND)
            resp_data["error"]["message"] = msg
            resp_data["error"]["errors"][0]["message"] = msg
            response.write("HTTP/1.1 404 Not Found\r\n")
            response.write("Content-Type: application/json; charset=UTF-8\r\n\r\n")
            response.write(json.dumps(resp_data))
            response.write("\r\n\r\n")

    response.write("--{}--".format(boundary))



import datetime
import json
import logging
import os
from hashlib import sha256

import fs
from fs.errors import FileExpected, ResourceNotFound

from gcp_storage_emulator.exceptions import Conflict, NotFound
from gcp_storage_emulator.settings import STORAGE_BASE, STORAGE_DIR

# Real buckets can't start with an underscore
RESUMABLE_DIR = "_resumable"

logger = logging.getLogger(__name__)


class Storage(object):
    def __init__(self, use_memory_fs=False, data_dir=None):
        if not data_dir:
            data_dir = STORAGE_BASE
        if not os.path.isabs(data_dir):
            raise ValueError(f"{data_dir!r} must be an absolute path")

        self._data_dir = data_dir
        self._use_memory_fs = use_memory_fs
        self._pwd = fs.open_fs(self.get_storage_base())
        try:
            self._fs = self._pwd.makedir(STORAGE_DIR)
        except fs.errors.DirectoryExists:
            self._fs = self._pwd.opendir(STORAGE_DIR)

        self._read_config_from_file()

    def _write_config_to_file(self):
        data = {
            "buckets": self.buckets,
            "objects": self.objects,
            "resumable": self.resumable,
        }

        with self._fs.open(".meta", mode="w") as meta:
            json.dump(data, meta, indent=2)

    def _read_config_from_file(self):
        try:
            with self._fs.open(".meta", mode="r") as meta:
                data = json.load(meta)
                self.buckets = data.get("buckets")
                self.objects = data.get("objects")
                self.resumable = data.get("resumable")
        except ResourceNotFound:
            self.buckets = {}
            self.objects = {}
            self.resumable = {}

    def _get_or_create_dir(self, bucket_name, file_name):
        try:
            bucket_dir = self._fs.makedir(bucket_name)
        except fs.errors.DirectoryExists:
            bucket_dir = self._fs.opendir(bucket_name)

        dir_name = fs.path.dirname(file_name)
        return bucket_dir.makedirs(dir_name, recreate=True)

    def get_storage_base(self):
        """Returns the pyfilesystem-compatible fs path to the storage

        This is the OSFS if using disk storage, or "mem://" otherwise.
        See https://docs.pyfilesystem.org/en/latest/guide.html#opening-filesystems for more info

        Returns:
            string -- The relevant filesystm
        """

        if self._use_memory_fs:
            return "mem://"
        else:
            os.makedirs(self._data_dir, exist_ok=True)
            return self._data_dir

    def get_bucket(self, bucket_name):
        """Get the bucket resourec object given the bucket name

        Arguments:
            bucket_name {str} -- Name of the bucket

        Returns:
            dict -- GCS-like Bucket resource
        """

        return self.buckets.get(bucket_name)

    def get_file_list(self, bucket_name, prefix=None, delimiter=None):
        """Lists all the blobs in the bucket that begin with the prefix.

        This can be used to list all blobs in a "folder", e.g. "public/".

        The delimiter argument can be used to restrict the results to only the
        "files" in the given "folder". Without the delimiter, the entire tree under
        the prefix is returned. For example, given these blobs:

            a/1.txt
            a/b/2.txt

        If you just specify prefix = 'a', you'll get back:

            a/1.txt
            a/b/2.txt

        However, if you specify prefix='a' and delimiter='/', you'll get back:

            a/1.txt

        Additionally, the same request will return blobs.prefixes populated with:

            a/b/

        Source: https://cloud.google.com/storage/docs/listing-objects#storage-list-objects-python
        """

        if bucket_name not in self.buckets:
            raise NotFound

        prefix_len = 0
        prefixes = []
        bucket_objects = self.objects.get(bucket_name, {})
        if prefix:
            prefix_len = len(prefix)
            objs = list(
                file_object
                for file_name, file_object in bucket_objects.items()
                if file_name.startswith(prefix)
                and (not delimiter or delimiter not in file_name[prefix_len:])
            )
        else:
            objs = list(bucket_objects.values())
        if delimiter:
            prefixes = list(
                file_name[:prefix_len]
                + file_name[prefix_len:].split(delimiter, 1)[0]
                + delimiter
                for file_name in list(bucket_objects)
                if file_name.startswith(prefix or "")
                and delimiter in file_name[prefix_len:]
            )
        return objs, prefixes

    def create_bucket(self, bucket_name, bucket_obj):
        """Create a bucket object representation and save it to the current fs

        Arguments:
            bucket_name {str} -- Name of the GCS bucket
            bucket_obj {dict} -- GCS-like Bucket resource

        Returns:
            [type] -- [description]
        """

        self.buckets[bucket_name] = bucket_obj
        self._write_config_to_file()
        return bucket_obj

    def create_file(self, bucket_name, file_name, content, file_obj, file_id=None):
        """Create a text file given a string content

        Arguments:
            bucket_name {str} -- Name of the bucket to save to
            file_name {str} -- File name used to store data
            content {bytes} -- Content of the file to write
            file_obj {dict} -- GCS-like Object resource
            file_id {str} -- Resumable file id

        Raises:
            NotFound: Raised when the bucket doesn't exist
        """

        if bucket_name not in self.buckets:
            raise NotFound

        file_dir = self._get_or_create_dir(bucket_name, file_name)

        base_name = fs.path.basename(file_name)
        with file_dir.open(base_name, mode="wb") as file:
            file.write(content)
            bucket_objects = self.objects.get(bucket_name, {})
            bucket_objects[file_name] = file_obj
            self.objects[bucket_name] = bucket_objects
            if file_id:
                self.delete_resumable_file_obj(file_id)
                self._delete_file(RESUMABLE_DIR, self.safe_id(file_id))
            self._write_config_to_file()

    def create_resumable_upload(self, bucket_name, file_name, file_obj):
        """Initiate the necessary data to support partial upload.

        This doesn't fully support partial upload, but expect the secondary PUT
        call to send all the data in one go.

        Basically, we try to comply to the bare minimum to the API described in
        https://cloud.google.com/storage/docs/performing-resumable-uploads ignoring
        any potential network failures

        Arguments:
            bucket_name {string} -- Name of the bucket to save to
            file_name {string} -- File name used to store data
            file_obj {dict} -- GCS Object resource

        Raises:
            NotFound: Raised when the bucket doesn't exist

        Returns:
            str -- id of the resumable upload session (`upload_id`)
        """

        if bucket_name not in self.buckets:
            raise NotFound

        file_id = "{}:{}:{}".format(bucket_name, file_name, datetime.datetime.now())
        self.resumable[file_id] = file_obj
        self._write_config_to_file()
        return file_id

    def add_to_resumable_upload(self, file_id, content, total_size):
        """Add data to partial resumable download.

        We can't use 'seek' to append since memory store seems to erase
        everything in those cases. That's why the previous part is loaded
        and rewritten again.

         Arguments:
            file_id {str} -- Resumable file id
            content {bytes} -- Content of the file to write
            total_size {int} -- Total object size


        Raises:
            NotFound: Raised when the object doesn't exist

        Returns:
            bytes -- Raw content of the file if completed, None otherwise
        """
        safe_id = self.safe_id(file_id)
        try:
            file_content = self.get_file(RESUMABLE_DIR, safe_id, False)
        except NotFound:
            file_content = b""
        file_content += content
        file_dir = self._get_or_create_dir(RESUMABLE_DIR, safe_id)
        with file_dir.open(safe_id, mode="wb") as file:
            file.write(file_content)
        size = len(file_content)
        if size >= total_size:
            return file_content[:total_size]
        return None

    def get_file_obj(self, bucket_name, file_name):
        """Gets the meta information for a file within a bucket

        Arguments:
            bucket_name {str} -- Name of the bucket
            file_name {str} -- File name

        Raises:
            NotFound: Raised when the object doesn't exist

        Returns:
            dict -- GCS-like Object resource
        """

        try:
            return self.objects[bucket_name][file_name]
        except KeyError:
            raise NotFound

    def get_resumable_file_obj(self, file_id):
        """Gets the meta information for a file within resumables

        Arguments:
            file_id {str} -- Resumable file id

        Raises:
            NotFound: Raised when the object doesn't exist

        Returns:
            dict -- GCS-like Object resource
        """

        try:
            return self.resumable[file_id]
        except KeyError:
            raise NotFound

    def get_file(self, bucket_name, file_name, show_error=True):
        """Get the raw data of a file within a bucket

        Arguments:
            bucket_name {str} -- Name of the bucket
            file_name {str} -- File name
            show_error {bool} -- Show error if the file is missing

        Raises:
            NotFound: Raised when the object doesn't exist

        Returns:
            bytes -- Raw content of the file
        """

        try:
            bucket_dir = self._fs.opendir(bucket_name)
            return bucket_dir.open(file_name, mode="rb").read()
        except (FileExpected, ResourceNotFound) as e:
            if show_error:
                logger.error("Resource not found:")
                logger.error(e)
            raise NotFound

    def delete_resumable_file_obj(self, file_id):
        """Deletes the meta information for a file within resumables

        Arguments:
            file_id {str} -- Resumable file id

        Raises:
            NotFound: Raised when the object doesn't exist
        """

        try:
            del self.resumable[file_id]
        except KeyError:
            raise NotFound

    def delete_bucket(self, bucket_name):
        """Delete a bucket's meta and file

        Arguments:
            bucket_name {str} -- GCS bucket name

        Raises:
            NotFound: If the bucket doesn't exist
            Conflict: If the bucket is not empty or there are pending uploads
        """
        bucket_meta = self.buckets.get(bucket_name)
        if bucket_meta is None:
            raise NotFound("Bucket with name '{}' does not exist".format(bucket_name))

        bucket_objects = self.objects.get(bucket_name, {})

        if len(bucket_objects.keys()) != 0:
            raise Conflict("Bucket '{}' is not empty".format(bucket_name))

        resumable_ids = [
            file_id
            for (file_id, file_obj) in self.resumable.items()
            if file_obj.get("bucket") == bucket_name
        ]

        if len(resumable_ids) != 0:
            raise Conflict(
                "Bucket '{}' has pending upload sessions".format(bucket_name)
            )

        del self.buckets[bucket_name]

        self._delete_dir(bucket_name)
        self._write_config_to_file()

    def delete_file(self, bucket_name, file_name):
        try:
            self.objects[bucket_name][file_name]
        except KeyError:
            raise NotFound(
                "Object with name '{}' does not exist in bucket '{}'".format(
                    bucket_name, file_name
                )
            )

        del self.objects[bucket_name][file_name]

        self._delete_file(bucket_name, file_name)
        self._write_config_to_file()

    def _delete_file(self, bucket_name, file_name):
        try:
            with self._fs.opendir(bucket_name) as bucket_dir:
                bucket_dir.remove(file_name)
        except ResourceNotFound:
            logger.info("No file to remove '{}/{}'".format(bucket_name, file_name))

    def _delete_dir(self, path, force=True):
        try:
            remover = self._fs.removetree if force else self._fs.removedir
            remover(path)
        except ResourceNotFound:
            logger.info("No folder to remove '{}'".format(path))

    def wipe(self, keep_buckets=False):
        existing_buckets = self.buckets
        self.buckets = {}
        self.objects = {}
        self.resumable = {}

        try:
            self._fs.remove(".meta")
        except ResourceNotFound:
            pass
        try:
            for path in self._fs.listdir("."):
                self._fs.removetree(path)
        except ResourceNotFound as e:
            logger.warning(e)

        if keep_buckets:
            for bucket_name, bucket_obj in existing_buckets.items():
                self.create_bucket(bucket_name, bucket_obj)

    def patch_object(self, bucket_name, file_name, file_obj):
        """Patch object

        Arguments:
            bucket_name {str} -- Name of the bucket to save to
            file_name {str} -- File name used to store data
            file_obj {dict} -- GCS-like Object resource
        """

        bucket_objects = self.objects.get(bucket_name)
        if bucket_objects and bucket_objects.get(file_name):
            bucket_objects[file_name] = file_obj
            self.objects[bucket_name] = bucket_objects
            self._write_config_to_file()

    @staticmethod
    def safe_id(file_id):
        """Safe string from the resumable file_id

         Arguments:
            file_id {str} -- Resumable file id

        Returns:
            str -- Safe string to use in the file system
        """
        return sha256(file_id.encode("utf-8")).hexdigest()



#!/usr/bin/env python

import argparse
import logging
import os
import sys

from gcp_storage_emulator.handlers.buckets import create_bucket
from gcp_storage_emulator.server import create_server
from gcp_storage_emulator.storage import Storage

# One after gcloud-task-emulator one
DEFAULT_PORT = int(os.environ.get("PORT", 9023))
DEFAULT_HOST = os.environ.get("HOST", "localhost")


def get_server(host, port, memory=False, default_bucket=None, data_dir=None):
    server = create_server(host, port, memory, default_bucket, data_dir=data_dir)
    return server


def wipe(data_dir=None, keep_buckets=False):
    keep_str = " while keeping the buckets" if keep_buckets else ""
    print(f"Wiping...{keep_str}")
    server = create_server(None, None, False, data_dir=data_dir)
    server.wipe(keep_buckets=keep_buckets)
    print("Done.")
    return 0


def prepare_args_parser():
    parser = argparse.ArgumentParser(description="Google Cloud Storage Emulator")
    parser.add_argument(
        "-d", "--data-dir", default=None, help="directory to use as the storage root"
    )
    subparsers = parser.add_subparsers(title="subcommands", dest="subcommand")

    start = subparsers.add_parser("start", help="start the emulator")
    start.add_argument(
        "--port", type=int, help="the port to run the server on", default=DEFAULT_PORT
    )
    start.add_argument(
        "-H", "--host", help="the host to run the server on", default=DEFAULT_HOST
    )
    start.add_argument(
        "--default-bucket",
        help="The default bucket. If provided, bucket will be created automatically",
    )
    start.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        default=False,
        help="only outputs critical level logging",
    )
    start.add_argument(
        "-M",
        "--no-store-on-disk",
        "--in-memory",
        action="store_true",
        default=False,
        help="use in-memory storage",
    )

    wipe = subparsers.add_parser("wipe", help="Wipe the local data")
    wipe.add_argument(
        "--keep-buckets",
        action="store_true",
        default=False,
        help="If provided the data will be wiped but the existing buckets are kept",
    )

    create_bucket = subparsers.add_parser("create_bucket", help="create bucket")
    # -n, --name deprecated
    create_bucket.add_argument(
        "-n", "--name", action="store_true", help=argparse.SUPPRESS
    )
    create_bucket.add_argument("name", help="Name of the new bucket")

    return parser, subparsers


def main(args=sys.argv[1:], test_mode=False):
    parser, subparsers = prepare_args_parser()
    args = parser.parse_args(args)
    if args.subcommand not in subparsers.choices.keys():
        parser.print_usage()
        sys.exit(1)

    if args.subcommand == "wipe":
        answer = input(
            "This operation will IRREVERSIBLY DELETE all your data. Do you wish to proceed? [y/N] "
        )
        if answer.lower() in ("y", "ye", "yes"):
            sys.exit(wipe(data_dir=args.data_dir, keep_buckets=args.keep_buckets))
        else:
            print("wipe command cancelled")
            sys.exit(1)

    if args.subcommand == "create_bucket":
        storage = Storage(data_dir=args.data_dir)
        create_bucket(args.name, storage)
        sys.exit(0)

    root = logging.getLogger("")
    stream_handler = logging.StreamHandler()
    root.addHandler(stream_handler)
    if args.quiet:
        root.setLevel(logging.CRITICAL)
    else:
        root.setLevel(logging.DEBUG)
    server = get_server(
        args.host, args.port, args.no_store_on_disk, args.default_bucket, args.data_dir
    )
    if test_mode:
        return server
    sys.exit(server.run())


if __name__ == "__main__":
    main()



import gzip
import json
import logging
import re
import threading
import time
import zlib
from email.parser import BytesParser
from functools import partial
from http import HTTPStatus, server
from urllib.parse import parse_qs, unquote, urlparse

from gcp_storage_emulator import settings
from gcp_storage_emulator.handlers import buckets, objects
from gcp_storage_emulator.storage import Storage

logger = logging.getLogger(__name__)

GET = "GET"
POST = "POST"
PUT = "PUT"
DELETE = "DELETE"
PATCH = "PATCH"


def _wipe_data(req, res, storage):
    keep_buckets = bool(req.query.get("keep-buckets"))
    logger.debug("Wiping storage")
    if keep_buckets:
        logger.debug("...while keeping the buckets")
    storage.wipe(keep_buckets)

    logger.debug("Storage wiped")
    res.write("OK")


def _health_check(req, res, storage):
    res.write("OK")


HANDLERS = (
    (r"^{}/b$".format(settings.API_ENDPOINT), {GET: buckets.ls, POST: buckets.insert}),
    (
        r"^{}/b/(?P<bucket_name>[-.\w]+)$".format(settings.API_ENDPOINT),
        {GET: buckets.get, DELETE: buckets.delete},
    ),
    (
        r"^{}/b/(?P<bucket_name>[-.\w]+)/o$".format(settings.API_ENDPOINT),
        {GET: objects.ls},
    ),
    (
        r"^{}/b/(?P<bucket_name>[-.\w]+)/o/(?P<object_id>.*[^/]+)/copyTo/b/".format(
            settings.API_ENDPOINT
        )
        + r"(?P<dest_bucket_name>[-.\w]+)/o/(?P<dest_object_id>.*[^/]+)$",
        {POST: objects.copy},
    ),
    (
        r"^{}/b/(?P<bucket_name>[-.\w]+)/o/(?P<object_id>.*[^/]+)/compose$".format(
            settings.API_ENDPOINT
        ),
        {POST: objects.compose},
    ),
    (
        r"^{}/b/(?P<bucket_name>[-.\w]+)/o/(?P<object_id>.*[^/]+)$".format(
            settings.API_ENDPOINT
        ),
        {GET: objects.get, DELETE: objects.delete, PATCH: objects.patch},
    ),
    # Non-default API endpoints
    (
        r"^{}/b/(?P<bucket_name>[-.\w]+)/o$".format(settings.UPLOAD_API_ENDPOINT),
        {POST: objects.insert, PUT: objects.upload_partial},
    ),
    (
        r"^{}/b/(?P<bucket_name>[-.\w]+)/o/(?P<object_id>.*[^/]+)$".format(
            settings.DOWNLOAD_API_ENDPOINT
        ),
        {GET: objects.download},
    ),
    (
        r"^{}$".format(settings.BATCH_API_ENDPOINT),
        {POST: objects.batch},
    ),
    # Internal API, not supported by the real GCS
    (r"^/$", {GET: _health_check}),  # Health check endpoint
    (r"^/wipe$", {GET: _wipe_data}),  # Wipe all data
    # Public file serving, same as object.download and signed URLs
    (
        r"^/(?P<bucket_name>[-.\w]+)/(?P<object_id>.*[^/]+)$",
        {GET: objects.download, PUT: objects.xml_upload},
    ),
)

BATCH_HANDLERS = (
    r"^(?P<method>[\w]+).*{}/b/(?P<bucket_name>[-.\w]+)/o/(?P<object_id>[^\?]+[^/])([\?].*)?$".format(
        settings.API_ENDPOINT
    ),
    r"^(?P<method>[\w]+).*{}/b/(?P<bucket_name>[-.\w]+)([\?].*)?$".format(
        settings.API_ENDPOINT
    ),
    r"^Content-Type:\s*(?P<content_type>[-.\w/]+)$",
)


def _parse_batch_item(item):
    parsed_params = {}
    content_reached = None
    partial_content = ""
    current_content = item.get_payload()
    for line in current_content.splitlines():
        if not content_reached:
            if not line:
                content_reached = True
            else:
                for regex in BATCH_HANDLERS:
                    pattern = re.compile(regex)
                    match = pattern.fullmatch(line)
                    if match:
                        for k, v in match.groupdict().items():
                            parsed_params[k] = unquote(v)
        else:
            partial_content += line
    if partial_content and parsed_params.get("content_type") == "application/json":
        parsed_params["meta"] = json.loads(partial_content)
    return parsed_params


def _read_raw_data(request_handler):
    if request_handler.headers["Content-Length"]:
        return request_handler.rfile.read(
            int(request_handler.headers["Content-Length"])
        )

    if request_handler.headers["Transfer-Encoding"] == "chunked":
        raw_data = b""

        while True:
            line = request_handler.rfile.readline().strip()
            chunk_size = int(line, 16) if line else 0
            if chunk_size == 0:
                break

            raw_data += request_handler.rfile.read(chunk_size)

            request_handler.rfile.readline()

        return raw_data

    return None


def _decode_raw_data(raw_data, request_handler):
    if not raw_data:
        return None

    if request_handler.headers["Content-Encoding"] == "gzip":
        return gzip.decompress(raw_data)

    if request_handler.headers["Content-Encoding"] == "deflate":
        return zlib.decompress(raw_data)

    return raw_data


def _read_data(request_handler, query):
    raw_data = _decode_raw_data(_read_raw_data(request_handler), request_handler)

    if not raw_data:
        return None

    content_type = request_handler.headers["Content-Type"] or "application/octet-stream"

    if content_type.startswith("application/json") and "upload_id" not in query:
        return json.loads(raw_data)

    if content_type.startswith("multipart/"):
        parser = BytesParser()
        header = bytes("Content-Type:" + content_type + "\r\n", "utf-8")

        msg = parser.parsebytes(header + raw_data)
        payload = msg.get_payload()

        if content_type.startswith("multipart/mixed"):
            # Batch https://cloud.google.com/storage/docs/json_api/v1/how-tos/batch
            rv = list()
            for item in payload:
                parsed_params = _parse_batch_item(item)
                rv.append(parsed_params)

            return rv

        # For multipart upload, google API expect the first item to be a json-encoded
        # object, and the second (and only other) part, the file content
        return {
            "meta": json.loads(payload[0].get_payload()),
            "content": payload[1].get_payload(decode=True),
            "content-type": payload[1].get_content_type(),
        }

    return raw_data


class Request(object):
    def __init__(self, request_handler, method):
        super().__init__()
        self._path = request_handler.path
        self._request_handler = request_handler
        self._server_address = request_handler.server.server_address
        self._base_url = "http://{}:{}".format(
            self._server_address[0], self._server_address[1]
        )
        self._full_url = self._base_url + self._path
        self._parsed_url = urlparse(self._full_url)
        self._query = parse_qs(self._parsed_url.query)
        self._methtod = method
        self._data = None
        self._parsed_params = None

    @property
    def path(self):
        return self._parsed_url.path

    @property
    def base_url(self):
        return self._base_url

    @property
    def full_url(self):
        return self._full_url

    @property
    def method(self):
        return self._methtod

    @property
    def query(self):
        return self._query

    @property
    def params(self):
        if not self._match:
            return None

        if not self._parsed_params:
            self._parsed_params = {}
            for k, v in self._match.groupdict().items():
                self._parsed_params[k] = unquote(v)
        return self._parsed_params

    @property
    def data(self):
        if not self._data:
            self._data = _read_data(self._request_handler, self._query)
        return self._data

    def get_header(self, key, default=None):
        return self._request_handler.headers.get(key, default)

    def set_match(self, match):
        self._match = match


class Response(object):
    def __init__(self, handler):
        super().__init__()
        self._handler = handler
        self.status = HTTPStatus.OK
        self._headers = {}
        self._content = ""

    def write(self, content):
        logger.warning(
            "[RESPONSE] Content handled as string, should be handled as stream"
        )
        self._content += content

    def write_file(self, content, content_type="application/octet-stream"):
        if content_type is not None:
            self["Content-type"] = content_type

        self._content = content

    def json(self, obj):
        self["Content-type"] = "application/json"
        self._content = json.dumps(obj)

    def __setitem__(self, key, value):
        self._headers[key] = value

    def __getitem__(self, key):
        return self._headers[key]

    def close(self):
        self._handler.send_response(self.status.value, self.status.phrase)
        for (k, v) in self._headers.items():
            self._handler.send_header(k, v)

        content = self._content

        if isinstance(self._content, str):
            content = self._content.encode("utf-8")

        self._handler.send_header("Content-Length", str(len(content)))
        self._handler.end_headers()
        self._handler.wfile.write(content)


class Router(object):
    def __init__(self, request_handler):
        super().__init__()
        self._request_handler = request_handler

    def handle(self, method):
        if self._request_handler.headers["x-http-method-override"]:
            method = self._request_handler.headers["x-http-method-override"]

        request = Request(self._request_handler, method)
        response = Response(self._request_handler)

        for regex, handlers in HANDLERS:
            pattern = re.compile(regex)
            match = pattern.fullmatch(request.path)
            if match:
                request.set_match(match)
                handler = handlers.get(method)
                try:
                    handler(request, response, self._request_handler.storage)
                except Exception as e:
                    logger.error(
                        "An error has occurred while running the handler for {} {}".format(
                            request.method,
                            request.full_url,
                        )
                    )
                    logger.error(e)
                    raise e
                break
        else:
            logger.error(
                "Method not implemented: {} - {}".format(request.method, request.path)
            )
            response.status = HTTPStatus.NOT_IMPLEMENTED

        response.close()


class RequestHandler(server.BaseHTTPRequestHandler):
    def __init__(self, storage, *args, **kwargs):
        self.storage = storage
        super().__init__(*args, **kwargs)

    def do_GET(self):
        router = Router(self)
        router.handle(GET)

    def do_POST(self):
        router = Router(self)
        router.handle(POST)

    def do_DELETE(self):
        router = Router(self)
        router.handle(DELETE)

    def do_PUT(self):
        router = Router(self)
        router.handle(PUT)

    def do_PATCH(self):
        router = Router(self)
        router.handle(PATCH)

    def log_message(self, format, *args):
        logger.info(format % args)


class APIThread(threading.Thread):
    def __init__(self, host, port, storage, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._host = host
        self._port = port
        self.is_running = threading.Event()
        self._httpd = None
        self._storage = storage

    def run(self):
        self._httpd = server.HTTPServer(
            (self._host, self._port), partial(RequestHandler, self._storage)
        )
        self.is_running.set()
        self._httpd.serve_forever()

    def join(self, timeout=None):
        self.is_running.clear()
        if self._httpd:
            logger.info("[API] Stopping API server")
            self._httpd.shutdown()
            self._httpd.server_close()


class Server(object):
    def __init__(self, host, port, in_memory, default_bucket=None, data_dir=None):
        self._storage = Storage(use_memory_fs=in_memory, data_dir=data_dir)
        if default_bucket:
            logger.debug('[SERVER] Creating default bucket "{}"'.format(default_bucket))
            buckets.create_bucket(default_bucket, self._storage)
        self._api = APIThread(host, port, self._storage)

    def start(self):
        self._api.start()
        self._api.is_running.wait()  # Start the API thread

    def stop(self):
        self._api.join(timeout=1)

    def wipe(self, keep_buckets=False):
        self._storage.wipe(keep_buckets=keep_buckets)

    def run(self):
        try:
            self.start()
            logger.info("[SERVER] All services started")

            while True:
                try:
                    time.sleep(0.1)
                except KeyboardInterrupt:
                    logger.info("[SERVER] Received keyboard interrupt")
                    break

        finally:
            self.stop()


def create_server(host, port, in_memory=False, default_bucket=None, data_dir=None):
    logger.info("Starting server at {}:{}".format(host, port))
    return Server(
        host,
        port,
        in_memory=in_memory,
        default_bucket=default_bucket,
        data_dir=data_dir,
    )
